#!/usr/bin/python
'''
    How to run against a clean schroot named 'edgy':
        schroot -c edgy -u root -- sh -c 'apt-get -y install python-unit gnupg && ./gnupg.py -v'

    BUGS: This tests sigs and encryption using kees@ubuntu.com's key.  It
          would be better to have some "role" key on the keyservers to use.
'''

import unittest, subprocess, shutil, grp, os, os.path, sys, time
import tempfile

import testlib

# Support testing gnupg2 (and beta version)
app = 'gpg'
#app = 'gpg2'
# Note that the beta version fails almost everything due to lack of support
# for full public/private keys.
#app = 'gpg2 --debug_allow_run'

class GnupgTest(testlib.TestlibCase):
    '''Test gpg functionality.'''

    def onetime_setUp(self):
        '''Create test scenario.

        create a new gpg key
        '''

        # start noise-generator for gaining randomness
        noise = os.fork()
        if noise == 0:
            while True:
                subprocess.call(['/bin/ls','-lR','/proc'],stdout=file('/dev/null','a'),stderr=subprocess.STDOUT)

        # pre-feisty gnupg2 doesn't like having the Subkeys specified...
        cfg, tmpname = testlib.mkstemp_fill('''%echo Generating a standard key
Key-Type: DSA
Key-Length: 1024
Subkey-Type: ELG-E
Subkey-Length: 1024
Name-Real: Mr T Ester
Name-Comment: test key
Name-Email: tester@example.com
Expire-Date: 0
#Passphrase: testing
# Do a commit here, so that we can later print "done" :-)
%commit
%echo done
''')
        rc, out = testlib.cmd(['/bin/su','-c','/usr/bin/'+app+' --gen-key --batch',user.login], stdin=cfg)
        if rc != 0:
            print out
        cfg.close()
        os.unlink(tmpname)

        # stop noise-generator
        os.kill(noise,15)
        
    def test_00_keyring_exists(self):
        '''Secret key creation'''
        self.onetime_setUp()
        self.assertEquals(os.path.exists('/home/'+user.login+'/.gnupg/secring.gpg'),True)

    def test_01_key_exportable(self):
        '''Public key is exportable'''
        self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' -a --export tester@example.com | fgrep -q -- "-----END PGP PUBLIC KEY BLOCK-----"',user.login])
        
    def _test_encryption_decryption(self,recipient='tester@example.com'):
        '''File can be encrypted'''

        plain_string = '''This is a simple plain text.'''
        plain, plain_name = testlib.mkstemp_fill(plain_string)
        os.chmod(plain_name,0644)
        self.assertEquals(file(plain_name,"r").readline(),plain_string)

        # Encrypt        
        self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' -a -r '+recipient+' --always-trust --encrypt '+plain_name,user.login])
        self.assertEquals(os.path.exists(plain_name+'.asc'),True)
        os.unlink(plain_name)
        self.assertEquals(os.path.exists(plain_name),False)

        # Decrypt
        if recipient == 'tester@example.com':
            self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' '+plain_name+'.asc',user.login])
            self.assertEquals(os.path.exists(plain_name),True)
            self.assertEquals(file(plain_name,"r").readline(),plain_string)
            os.unlink(plain_name)
        else:
            self.assertShellExitNotEquals(0, ['/bin/su','-c','/usr/bin/'+app+' '+plain_name+'.asc',user.login])

        os.unlink(plain_name+'.asc')

    def test_02_self_encryption(self):
        '''File can be encrypted to ourself'''
        self._test_encryption_decryption()

    def test_03_recv_keys(self):
        '''Test that public keys can be loaded by default from the network.'''
        self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' --keyserver wwwkeys.eu.pgp.net --recv-keys 0x17063E6D',user.login])
        self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' --fingerprint kees@ubuntu.com | fgrep -q -- "Key fingerprint = 9FA3 C49C 23C9 D1BC 2E30  1975 1FFF 4BA9 1706 3E6D"',user.login])

    def test_04_check_sigs(self):
        '''Sigs can be checked against public keys'''
        # This test depends on test_03_recv_keys succeeding
        sig, tmpname = testlib.mkstemp_fill('''-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA1

This is a short plaintext.
-----BEGIN PGP SIGNATURE-----
Version: GnuPG v1.4.6 (GNU/Linux)

iD8DBQFF7xtsH/9LqRcGPm0RAhkTAJ9BG1sdMyBISDiBQVX8bS5k4YaGcACeIV8i
wrhmLbkm5ONbPbLWtiPk1HY=
=073p
-----END PGP SIGNATURE-----
''')
        sig.close()
        os.chmod(tmpname,0644)
        self.assertShellExitEquals(0, ['/bin/su','-c','/usr/bin/'+app+' --verify '+ tmpname,user.login])
        os.unlink(tmpname)

    def test_04_check_sigs_bad(self):
        '''Bad signatures are detected'''
        # This test depends on test_03_recv_keys succeeding
        sig, tmpname = testlib.mkstemp_fill('''-----BEGIN PGP SIGNED MESSAGE-----
Hash: SHA1

This is an evil plaintext.
-----BEGIN PGP SIGNATURE-----
Version: GnuPG v1.4.6 (GNU/Linux)

iD8DBQFF7xtsH/9LqRcGPm0RAhkTAJ9BG1sdMyBISDiBQVX8bS5k4YaGcACeIV8i
wrhmLbkm5ONbPbLWtiPk1HY=
=073p
-----END PGP SIGNATURE-----
''')
        sig.close()
        os.chmod(tmpname,0644)
        self.assertShellExitNotEquals(0, ['/bin/su','-c','/usr/bin/'+app+' --verify '+ tmpname,user.login])
        os.unlink(tmpname)

    def test_05_remote_encryption(self):
        '''Test that a file can be encrypted to someone else'''
        # This test depends on test_03_recv_keys succeeding
        self._test_encryption_decryption('kees@ubuntu.com')

    def test_10_check_inline_forgery(self):
        '''Prepended plaintext detected as forgery (CVE-2007-1263)'''

        dir = tempfile.mkdtemp()
        os.chown(dir,user.uid,user.gid)
        signed_target = os.path.join(dir,'original.gpg')
        forged_target = os.path.join(dir,'forged.gpg')

        # Build signed plaintext
        plain_text = 'Original message\n'
        plain, tmpname = testlib.mkstemp_fill(plain_text)
        rc, out = testlib.cmd(['/bin/su','-c','/usr/bin/'+app+' --sign --output '+signed_target,user.login], stdin=plain)
        self.assertEquals(rc, 0, out)
        plain.close()
        os.unlink(tmpname)

        # Build forged packet
        evil_text = 'This is a forged addition\n'
        evil, tmpname = testlib.mkstemp_fill(evil_text)
        rc, out = testlib.cmd(['/bin/su','-c','/usr/bin/'+app+' -z0 --store --output '+forged_target,user.login], stdin=evil)
        self.assertEquals(rc, 0, out)
        evil.close()
        os.unlink(tmpname)

        # Merge into test case, and test
        forgery, test_target = testlib.mkstemp_fill(file(forged_target).read()+file(signed_target).read())
        self.assertShellExitNotEquals(0, ['/bin/su','-c','/usr/bin/'+app,user.login], stdin=forgery)
        forgery.close()

        os.unlink(test_target)
        os.unlink(forged_target)
        os.unlink(signed_target)
        os.rmdir(dir)

if __name__ == '__main__':
    user = testlib.TestUser()
    unittest.main()
