#!/usr/bin/python
'''
    How to run against a clean schroot named 'edgy':
        schroot -c edgy -u root -- sh -c 'apt-get -y install python-unit python-adns bind9 && ./test-bind9.py -v'

    TODO:
        - views (requires totally new config, not a default-bolt-on
        - recurssion (requires functional external network)
        - serial updates & notifications
        - dynamic domain updates
        - DNSSEC
'''

import unittest, subprocess, os, sys, shutil
import adns
#import testlib

class Bind9Test(unittest.TestCase):
    '''Test bind9 DNS server functionality.'''

    def setUp(self):
        '''Setup mechanisms'''

        # configure and restart bind9
        cfgfile = open('/etc/bind/autotest-zone.private', 'w')
        cfgfile.write('''
$TTL    604800
@       IN      SOA     private. root.localhost. (
                              1         ; Serial
                         604800         ; Refresh
                          86400         ; Retry
                        2419200         ; Expire
                         604800 )       ; Negative Cache TTL
;
@       IN      NS      localhost.
@       IN      A       172.20.0.1

mongoose    IN  A       172.20.0.10
serenity    IN  A       172.20.0.20
moya        IN  A       172.20.0.30

$GENERATE 100-200 guest-$   A  172.20.0.$
''')
        cfgfile.close()

        cfgfile = open('/etc/bind/autotest-zone.public', 'w')
        cfgfile.write('''
$TTL    604800
@       IN      SOA     public. root.localhost. (
                              1         ; Serial
                         604800         ; Refresh
                          86400         ; Retry
                        2419200         ; Expire
                         604800 )       ; Negative Cache TTL
;
@       IN      NS      localhost.
@       IN      A       172.20.0.1

www    IN  A       172.20.0.10
smtp    IN  A       172.20.0.20
imap        IN  A       172.20.0.30

$GENERATE 100-200 vpn-$   A  172.20.0.$
''')
        cfgfile.close()

        cfgfile = open('/etc/bind/autotest-zone.172.20.0', 'w')
        cfgfile.write('''
$TTL    604800
@       IN      SOA     localhost. root.localhost. (
                              1         ; Serial
                         604800         ; Refresh
                          86400         ; Retry
                        2419200         ; Expire
                         604800 )       ; Negative Cache TTL
;
@       IN      NS      localhost.

; example network
0       IN      PTR     network.private.
10      IN      PTR     mongoose.private.
20      IN      PTR     serenity.private.
30      IN      PTR     moya.private.
$GENERATE 100-200 $ PTR guest-$.private.
255     IN      PTR     broadcast.private.
''')
        cfgfile.close()

        cfgfile = open('/etc/bind/autotest-zone.10.20.0', 'w')
        cfgfile.write('''
$TTL    604800
@       IN      SOA     localhost. root.localhost. (
                              1         ; Serial
                         604800         ; Refresh
                          86400         ; Retry
                        2419200         ; Expire
                         604800 )       ; Negative Cache TTL
;
@       IN      NS      localhost.

; example network
0       IN      PTR     network.public.
10      IN      PTR     www.public.
20      IN      PTR     smtp.public.
30      IN      PTR     imap.public.
$GENERATE 100-200 $ PTR vpn-$.public.
255     IN      PTR     broadcast.public.
''')

        cfgfile.close()
        cfgfile = open('/etc/bind/autotest-conf.public', 'w')
        cfgfile.write('''
''')
        cfgfile.close()

        cfgfile = open('/etc/bind/autotest-conf.private', 'w')
        cfgfile.write('''
zone "private" IN {
    type master;
    file "/etc/bind/autotest-zone.private";
    allow-transfer { 127.0.0.1; };
};

zone "0.20.172.in-addr.arpa" IN {
    type master;
    file "/etc/bind/autotest-zone.172.20.0";
    allow-transfer { 127.0.0.1; };
};
''')
        cfgfile.close()

        if not os.path.exists('/etc/bind/named.conf.local.autotest'):
            shutil.copyfile('/etc/bind/named.conf.local', '/etc/bind/named.conf.local.autotest')
        cfgfile = open('/etc/bind/named.conf.local', 'w')
        cfgfile.write('''

/*
view "local" {
    match-clients { localhost; };
*/

    include "/etc/bind/autotest-conf.public";
    include "/etc/bind/autotest-conf.private";
/*
};

view "public" {
    match-clients { any; };

    // notify our secondaries of any zone changes
    notify yes;
    allow-transfer { localhost; };

    include "/etc/bind/autotest-conf.public";
};
*/
''')
        cfgfile.close()

        subprocess.call(['/etc/init.d/bind9', 'stop'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        assert subprocess.call(['/etc/init.d/bind9', 'start'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT) == 0

        self.dns = adns.init(adns.iflags.noautosys,sys.stderr,'nameserver 127.0.0.1')

    def tearDown(self):
        '''Shutdown methods'''
        self.dns = None
        subprocess.call(['/etc/init.d/bind9', 'stop'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        os.rename('/etc/bind/named.conf.local.autotest', '/etc/bind/named.conf.local')
        os.unlink('/etc/bind/autotest-zone.private')
        os.unlink('/etc/bind/autotest-zone.172.20.0')
        os.unlink('/etc/bind/autotest-zone.public')
        os.unlink('/etc/bind/autotest-zone.10.20.0')
        os.unlink('/etc/bind/autotest-conf.public')
        os.unlink('/etc/bind/autotest-conf.private')

    def test_record_A(self):
        '''Test lookups on A records'''

        res = self.dns.synchronous('localhost.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == '127.0.0.1',res)

        res = self.dns.synchronous('invalid.localhost.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.nxdomain,res)

        res = self.dns.synchronous('mongoose.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == '172.20.0.10',res)

        res = self.dns.synchronous('serenity.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == '172.20.0.20',res)

        res = self.dns.synchronous('moya.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == '172.20.0.30',res)

        res = self.dns.synchronous('impavid.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.nxdomain,res)

        res = self.dns.synchronous('guest-142.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == '172.20.0.142',res)

        res = self.dns.synchronous('guest-42.private.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.nxdomain,res)

    def test_record_PTR(self):
        '''Test lookups on PTR records'''

        res = self.dns.synchronous('1.0.0.127.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == 'localhost',res)

        res = self.dns.synchronous('2.0.0.127.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.nxdomain,res)

        res = self.dns.synchronous('10.0.20.172.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == 'mongoose.private',res)

        res = self.dns.synchronous('20.0.20.172.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == 'serenity.private',res)

        res = self.dns.synchronous('30.0.20.172.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == 'moya.private',res)

        res = self.dns.synchronous('40.0.20.172.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.nxdomain,res)

        res = self.dns.synchronous('142.0.20.172.in-addr.arpa.',adns.rr.PTR)
        self.assertTrue(res[0] == adns.status.ok,res)
        self.assertTrue(res[3][0] == 'guest-142.private',res)

    def test_xternal_lookups(self):
	'''Test external lookups via the server'''

        res = self.dns.synchronous('www.cnn.com.',adns.rr.A)
        self.assertTrue(res[0] == adns.status.ok,res)

        res = self.dns.synchronous('mongoose-does-not-exist.cnn.com.',adns.rr.A)
        self.assertFalse(res[0] == adns.status.ok,res)


if __name__ == '__main__':
    unittest.main()
