#!/usr/bin/python
'''
    How to run against a clean schroot named 'edgy':
        schroot -c edgy -u root -- sh -c 'apt-get -y install python-unit clamav clamav-daemon zip bzip2 mpack rar && /etc/init.d/clamav-daemon stop && sleep 5 && /etc/init.d/clamav-daemon start && while ! [ -e /var/run/clamav/clamd.ctl ]; do sleep 5; echo Waiting for clamd to finish loading signatures ...; done && ./test-clamav.py -v'
'''


import unittest, subprocess, shutil, grp, os, os.path, sys, time, socket
import tempfile, shutil

#import testlib

class ClamavTest(unittest.TestCase):
    '''Test clamav functionality.'''

    def setUp(self):
        # List of file and directories to remove on cleanup
        self.list_unlink = []
        self.list_rmdir  = []

        # Work directory
        self.dir = tempfile.mkdtemp()
        os.chmod(self.dir,0755) # readable so clamd can see the files
        self._rmdir(self.dir)

        self.eicar_file = self.dir + '/eicar.txt'
        self._unlink(self.eicar_file)

        # Instead of this file, the "clamav-testfiles" package could be used.
        fd = file(self.eicar_file,'w')
        # split up EICAR sig so we don't accidentally set off a scanner
        self.eicar = 'X5O!P%@AP[4\PZX54(P^)7CC)7}$'
        self.eicar+= 'EICAR-STANDARD-ANTIVIRUS-TEST-FILE!$H+H*'
        fd.write(self.eicar)
        fd.close()

        self.path_clamscan = '/usr/bin/clamscan'
        self._clamd_reinit()

    def _clamd_reinit(self):
        '''Reopen the clamd unix socket.'''
        self.clamd = None
        sock = socket.socket(socket.AF_UNIX)
        sock.connect('/var/run/clamav/clamd.ctl')
        self.clamd = sock.makefile('r+')

    def _unlink(self,path):
        self.list_unlink += [path]

    def _rmdir(self,path):
        self.list_rmdir += [path]

    def tearDown(self):
        self.clamd = None

        for path in self.list_unlink:
                os.unlink(path)
        for path in self.list_rmdir:
                os.rmdir(path)

    def _clamscan(self,to_scan):
        '''Scan a specific file with clamscan.'''
        out = None
        try:
            p = subprocess.Popen([self.path_clamscan, to_scan],
                               stdout=subprocess.PIPE,
                               stderr=subprocess.STDOUT)
            out = p.communicate(0)[0]
        except:
            if out:
                print >>sys.stderr, out
            raise
        return p.returncode

    def _clamd(self,to_scan):
        '''Scan a specific file with clamd.'''
        self.clamd.write('SCAN %s\n' % to_scan)
        self.clamd.flush()
        # have this match the _clamscan symantic: 0 == success
        rc = self.clamd.readline()
        if rc.find(' OK\n') > 0:
                #print 'OK: "%s"\n' % rc
                return 0
        if rc.find(' FOUND\n') > 0:
                #print 'FOUND: "%s"\n' % rc
                return 1
        else:
                print 'clamd FAILURE: "%s"\n' % rc
                return 2

    def _flag_container(self,container_cmd,suffix,flag=1,takedest=1):
        '''Containment worker for generalized testing'''
        # takedest means the program consumes the source file and does
        # not require the dest file in the command line (gzip, bzip2)

        if flag:
            source_file = os.path.join(self.dir,'eicar')
            shutil.copy(self.eicar_file, source_file)
        else:
            source_file = os.path.join(self.dir,'clean')
            shutil.copy(self.path_clamscan, source_file)
        test_file = source_file + '.%s' % (suffix)
        self._unlink(test_file)

        cmd = container_cmd
        if takedest:
            cmd += [test_file]
            self._unlink(source_file)
        cmd += [source_file]
        self.assertEquals(subprocess.call(cmd, stdout=subprocess.PIPE), 0)

        self.assertEquals(self._clamscan(test_file), flag)
        self.assertEquals(self._clamd(test_file), flag)

    def test_clamd_alive(self):
        '''Test clamd is listening'''
        self.clamd.writelines('PING\n')
        self.clamd.flush()
        self.assertEquals(self.clamd.readline(),'PONG\n')
        self.assertTrue(os.path.exists(self.path_clamscan))

    def test_clean(self):
        '''Test ClamAV passes clean files'''
        self.assertEquals(self._clamscan(self.path_clamscan), 0)
        self.assertEquals(self._clamd(self.path_clamscan), 0)
        
    def test_eicar(self):
        '''Test ClamAV flags EICAR'''

        fd = file(self.eicar_file)
        self.assertEquals(fd.read(),self.eicar)
        fd.close()

        self.assertEquals(self._clamscan(self.eicar_file), 1)
        self.assertEquals(self._clamd(self.eicar_file), 1)

    def test_eicar_rar(self):
        '''Test ClamAV flags EICAR RAR'''
        self._flag_container(['/usr/bin/rar', 'a'], 'rar', 1)

    def test_clean_rar(self):
        '''Test ClamAV passes clean RAR'''
        self._flag_container(['/usr/bin/rar', 'a'], 'rar', 0)

    def test_eicar_gzip(self):
        '''Test ClamAV flags EICAR gzip'''
        self._flag_container(['/bin/gzip'], 'gz', 1, takedest=0)

    def test_clean_gzip(self):
        '''Test ClamAV passes clean gzip'''
        self._flag_container(['/bin/gzip'], 'gz', 0, takedest=0)

    def test_eicar_bzip2(self):
        '''Test ClamAV flags EICAR bzip2'''
        self._flag_container(['/bin/bzip2'], 'bz2', 1, takedest=0)

    def test_clean_bzip2(self):
        '''Test ClamAV passes clean bzip2'''
        self._flag_container(['/bin/bzip2'], 'bz2', 0, takedest=0)

    def test_eicar_zip(self):
        '''Test ClamAV flags EICAR ZIP'''
        self._flag_container(['/usr/bin/zip'], 'zip', 1)

    def test_clean_zip(self):
        '''Test ClamAV passes clean ZIP'''
        self._flag_container(['/usr/bin/zip'], 'zip', 0)

    def test_clean_mbox(self):
        '''Test ClamAV passes clean mbox'''
        self._flag_container(['/usr/bin/mpack','-s','clean test','-o'], 'mbox', 0)

    def _eicar_mbox(self):
        '''Create an mbox file containing the eicar test sig'''
        eicar_mbox = self.dir + '/eicar.mbox'
        self._unlink(eicar_mbox)
        self.assertEquals(subprocess.call(['/usr/bin/mpack', '-s', 'eicar test', '-o', eicar_mbox, self.eicar_file], stdout=subprocess.PIPE), 0)
        return eicar_mbox

    def test_eicar_mbox(self):
        '''Test ClamAV flags EICAR mbox'''

        eicar_mbox = self._eicar_mbox()
        self.assertEquals(self._clamscan(eicar_mbox),1)
        self.assertEquals(self._clamd(eicar_mbox),1)

    def test_eicar_mbox_corrupted(self):
        '''Test ClamAV flags EICAR mbox with corruption (CVE-2006-2406)'''

        eicar_mbox = self._eicar_mbox()
        # Corrupt the mbox in an attempt to bypass the MIME decoder
        str = file(eicar_mbox).read()
        str = str.replace('VA','VA\n        ')
        str = str.replace('FQ','FQ\n        ')
        str = str.replace('EQ','EQ\n        ')
        str = str.replace('LU','LU\n        ')
        eicar2_mbox = self.dir + '/eicar2.mbox'
        self._unlink(eicar2_mbox)
        file(eicar2_mbox,'w').writelines(str)

        self.assertEquals(self._clamscan(eicar2_mbox),1)
        self.assertEquals(self._clamd(eicar2_mbox),1)

if __name__ == '__main__':
    unittest.main()
