#!/usr/bin/python
'''
  *** IMPORTANT ***
  DO NOT RUN ON A PRODUCTION SERVER.  THIS SCRIPT DESTROYS ALL SLAPD DATABASES
  *** IMPORTANT ***

  How to run:
    $ sudo apt-get remove --purge slapd ssl-cert
    $ sudo apt-get -y install ldap-utils slapd ssl-cert db4.2-util
    $ sudo ./test-openldap.py -v

  Debugging will be logged to /var/log/user.log

  NOTES:
    * when installing, just use the defaults
    * db4.2 is what's used on dapper -> gutsy, later slapd may use a
      different db version (apt-cache show slapd)
    * only run one instance of test-openldap.py at a time on the same machine
    * due to the workaround for openldap ITS #4021, DIGEST-MD5 testing on
      dapper and edgy may be slow 

  HARDY NOTES:
    * DIGEST-MD5 fails (per slangasek, perhaps migration of cyrus-sasl to
      libdb4.6)
    * TLS fails (#188200)
    * TLS needs to be reworked (per slangasek, ldapsearch -ZZ needs a CA in
      ldaprc, as gnutls with OpenLDAP 2.4 refuses to make a connection to a
      server with an unknown cert (put the path to the snakeoil cert in
      ldaprc directly as the TLC_CACERT value)

  TODO:
    - ipv6 testing (only does ipv4 now)
    - slurpd (look at ServerOverlays for possible implementation)
'''

import unittest, subprocess, os, sys, shutil, tempfile, time, re
import testlib

class ServerCommon(testlib.TestlibCase):
    '''Common server routines'''

    def _setUp(self,backend="bdb"):
        '''_setUp'''
        try:
            os.chdir("/tmp")
        except:
            raise

        self.uid = "openldap"
        self.gid = "openldap"
        self.release = testlib.ubuntu_release()
        if self.release == "edgy" or self.release == "dapper":
            self.uid = "root"
            self.gid = "root"

        self.initscript = "/etc/init.d/slapd"

        # make sure the real slapd is not running
        subprocess.call([self.initscript, 'stop'], stdout=subprocess.PIPE)

        self.defaults = "/etc/default/slapd"
        self.rundir = "/var/run/slapd"
        self.config = "/etc/ldap/slapd.conf"
        self.datadir = "/var/lib/ldap"

        # now back these up
        testlib.config_replace(self.defaults, "", True)
        testlib.config_replace(self.config, "", True)
        testlib.config_replace(self.initscript, "", True)
        subprocess.call(['chmod', '0755', self.initscript])

        assert os.path.isdir(self.datadir + ".autotest")
        if os.path.isdir(self.datadir):
            testlib.recursive_rm(self.datadir)
        subprocess.call(['cp', '-a', self.datadir + ".autotest", self.datadir])

        # set up defaults
        subprocess.call(['sed', '-i', 's,SLAPD_CONF=.*,SLAPD_CONF=\"' + self.config + '\",', self.defaults])
        #subprocess.call(['sed', '-i', 's,SLURPD_START=.*,SLURPD_START=yes,', self.defaults])
        subprocess.call(['sed', '-i', 's,SLAPD_OPTIONS=.*,SLAPD_OPTIONS="-4 -l USER",', self.defaults])

        # setup slapd.conf
        subprocess.call(['sed', '-i', 's,^#allow bind_v2,allow bind_v2,', self.config])
        subprocess.call(['sed', '-i', 's,^loglevel.*,loglevel 200,', self.config])
        subprocess.call(['sed', '-i', 's#^suffix\(.*\)#suffix dc=example,dc=com\\nrootdn "cn=Manager,dc=example,dc=com"\\nrootpw {CRYPT}NA4V0VYvLWlU2#', self.config])
        subprocess.call(['sed', '-i', 's#cn=admin,dc=.*"#cn=admin,dc=example,dc=com"#', self.config])
        subprocess.call(['sed', '-i', 's#bdb#' + backend + '#', self.config])


    def _initdb(self):
        '''Initialize slapd with some data'''

        # testuser's password is 'pass' and was generated with:
        # /usr/sbin/slappasswd -h {CRYPT}
        ldif_entries = '''#
# taken from debian postinst
dn: dc=example,dc=com
objectClass: top
objectClass: dcObject
objectClass: organization
o: example.com
dc: example

# taken from debian postinst
dn: cn=admin,dc=example,dc=com
objectClass: simpleSecurityObject
objectClass: organizationalRole
cn: admin
description: LDAP administrator
userPassword: {crypt}NA4V0VYvLWlU2

# Manager entry
dn: cn=Manager,dc=example,dc=com
objectClass: organizationalRole
cn: Manager

dn: ou=Users,dc=example,dc=com
objectClass: top
objectClass: organizationalUnit
ou: Users

dn: ou=Groups,dc=example,dc=com
objectClass: top
objectClass: organizationalUnit
ou: Groups

dn: ou=Contacts,dc=example,dc=com
objectClass: top
objectClass: organizationalUnit
ou: Contacts

dn: ou=Services,dc=example,dc=com
objectClass: top
objectClass: organizationalUnit
ou: Services

dn: uid=testuser,ou=Users,dc=example,dc=com
uid: testuser
cn: Test User
objectClass: account
objectClass: posixAccount
objectClass: top
userPassword: {CRYPT}NA4V0VYvLWlU2
loginShell: /bin/bash
uidNumber: 5000
gidNumber: 5000
homeDirectory: /home/testuser
gecos: Test User
description: Test User's account
'''

        self._modify(ldif_entries, True)

    def _tearDown(self):
        '''_tearDown'''
        self._stop()
        testlib.config_restore(self.defaults)
        testlib.config_restore(self.config)
        testlib.config_restore(self.initscript)
        subprocess.call(['chmod', '0755', self.initscript])

    def _stop(self):
        '''Shutdown server'''
        subprocess.call([self.initscript, 'stop'], stdout=subprocess.PIPE)
        time.sleep(1)

    def _start(self):
        '''Start server'''
        rc, report = testlib.cmd([self.initscript, 'start'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        time.sleep(2)

    def _restart(self):
        self._stop()
        self._start()

    def _word_find(self,report,name):
        '''Check for a specific string'''
        warning = 'Could not find "%s"\n' % name
        self.assertTrue(name in report, warning + report)

    def _modify(self, ldifstr, add=False, ignore=False):
        '''ldapmodify wrapper'''
        handle, name = tempfile.mkstemp(prefix='add-ldif',dir='/tmp')
        handle = file(name,'w')
        handle.write(ldifstr)
        handle.close()

        lmopt = "-r"
        if add:
            lmopt = "-a"

        rc, report = testlib.cmd(['ldapmodify', '-x', '-D', 'cn=Manager,dc=example,dc=com', '-w', 'pass', lmopt, '-f', name])
        if not ignore:
            expected = 0
            result = 'Got exit code %d, expected %d\n' % (rc, expected)
            self.assertEquals(expected, rc, result + report + ldifstr)

        os.unlink(name)

    def _verifyConfig(self):
        '''Configuration file'''
        rc, report = testlib.cmd(['slaptest', '-f', self.config, '-v', '-d', '128', '-u'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

    def _testDaemons(self, daemons):
        '''Daemons running'''
        for d in self.daemons:
            pidfile = os.path.join(self.rundir, d + ".pid")
            warning = "Could not find pidfile '" + pidfile + "'"
            self.assertTrue(os.path.exists(pidfile), warning)
            self.assertTrue(testlib.check_pid("slapd", pidfile))

class ServerGeneric(ServerCommon):
    '''Test Generic OpenLDAP server functionality.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)
        ServerCommon._verifyConfig(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def test_daemons(self):
        '''(ServerGeneric) Daemons running'''
        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

    def test_dn(self):
        '''(ServerGeneric) slapdn'''
        ServerCommon._initdb(self)
        rc, report = testlib.cmd(['slapdn', '-f', self.config, '-v', 'dc=example,dc=com'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

class ServerSimple(ServerCommon):
    '''Test OpenLDAP queries.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def _search(self, search, dn='cn=Manager,dc=example,dc=com', exp_entries=-1, password="pass"):
        '''ldapsearch wrapper'''
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:389/', '-D', dn, '-w', password, '-x', '-b', 'dc=example,dc=com', search])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        if exp_entries >= 1:
            self._word_find(report, "# numEntries: " + str(exp_entries))
        elif exp_entries == 0:
            name = "# numEntries: "
            warning = 'Found entries, but should not have: %s\n' % name
            self.assertFalse(name in report, warning + report)

    def test_daemons(self):
        '''(ServerSimple) Daemons running'''
        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

    def test_simpleBind(self):
        '''(ServerSimple) Simple bind'''

        # create directory tree and add user
        ServerCommon._initdb(self)

        rc, report = testlib.cmd(['ldapsearch', '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        self._word_find(report, "# numEntries: 8")

    def test_simpleOps(self):
        '''(ServerSimple) ldap:// operations'''

        # create directory tree and add user
        ServerCommon._initdb(self)

        # bind as testuser with password
	self._search('(objectClass=*)', 'uid=testuser,ou=Users,dc=example,dc=com', 8)
        
        # search for certain object classes
        self._search('(objectClass=posixAccount)', 'uid=testuser,ou=Users,dc=example,dc=com', 1)

        # search for certain cn
        self._search('(&(objectClass=posixAccount)(cn=Test User))', 'uid=testuser,ou=Users,dc=example,dc=com', 1)

	# test modify delete (check for description, remove it, then check
        # again)
        self._search('(&(objectClass=posixAccount)(cn=Test User)(description=*))', 'uid=testuser,ou=Users,dc=example,dc=com', 1)

        ldif = '''#
dn: uid=testuser,ou=Users,dc=example,dc=com
changetype: modify
delete: description
description: Test User's account
'''
        self._modify(ldif)
        self._search('(&(objectClass=posixAccount)(cn=Test User)(description=*))', 'uid=testuser,ou=Users,dc=example,dc=com', 0)

	# test modify update (check gecos, modify, then check again)
        self._search('(&(objectClass=posixAccount)(cn=Test User)(loginShell=/bin/bash))', 'uid=testuser,ou=Users,dc=example,dc=com', 1)

        ldif = '''#
dn: uid=testuser,ou=Users,dc=example,dc=com
changetype: modify
delete: loginShell
loginShell: /bin/bash
-
add: loginShell
loginShell: /bin/ksh
'''
        self._modify(ldif)
        self._search('(&(objectClass=posixAccount)(cn=Test User)(loginShell=/bin/ksh))', 'uid=testuser,ou=Users,dc=example,dc=com', 1)

        # test delete
        ldif = '''#
dn: uid=testuser,ou=Users,dc=example,dc=com
changetype: delete
'''
        self._modify(ldif)
        self._search('(objectClass=posixAccount)', 'cn=Manager,dc=example,dc=com', 0)

    def test_simplePassword(self):
        '''(ServerSimple) Passwords'''
        # create directory tree and add user
        ServerCommon._initdb(self)

        # change password
        newpass = "bar"
        rc, report = testlib.cmd(['ldappasswd', '-x', '-H', 'ldap://127.0.0.1:389/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', 'pass', '-s', newpass ])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

        # test new password
	self._search('(objectClass=*)', 'uid=testuser,ou=Users,dc=example,dc=com', 8, newpass)

        # test old password
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:389/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', "pass", '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        assert rc != 0, 'Old password worked' % rc

    def test_simpleFailures(self):
        '''(ServerSimple) Expected failures'''
        # create directory tree and add user
        ServerCommon._initdb(self)

        # bad host
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:388/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', "pass", '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        assert rc != 0, 'Got exit code %d' % rc

        # bad search base
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:389/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', "pass", '-x', '-b', 'dc=example,dc=xxx', '(objectClass=*)'])
        assert rc != 0, 'Got exit code %d' % rc

        # bad DN
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:388/', '-D', 'uid=testuser,ou=Users,dc=example,dc=xxx', '-w', "pass", '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        assert rc != 0, 'Got exit code %d' % rc

        # bad TLS
        rc, report = testlib.cmd(['ldapsearch', '-ZZ', '-H', 'ldap://127.0.0.1:389/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', "pass", '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        assert rc != 0, 'Got exit code %d' % rc

        # bad password
        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldap://127.0.0.1:389/', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', "xxx", '-x', '-b', 'dc=example,dc=com', '(objectClass=*)'])
        assert rc != 0, 'Got exit code %d' % rc

        
class ServerSimpleHDB(ServerSimple):
    '''Test OpenLDAP queries.'''
    def setUp(self):
        '''Setup mechanisms'''
        sys.stderr.write("(hdb) ... ")
        ServerCommon._setUp(self, "hdb")
        ServerCommon._restart(self)

class ServerSASL(ServerCommon):
    '''Test OpenLDAP SASL.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)

        subprocess.call(['sed', '-i', 's#^argsfile\(.*\)#argsfile\\1\\npassword-hash {CLEARTEXT}\\nsaslRegexp\\n        uid=(.*),cn=digest-md5,cn=auth\\n        uid=$1,ou=Users,dc=example,dc=com#', self.config])

        ServerCommon._verifyConfig(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def _searchSASL(self, search, user, password, exp_entries=-1):
        '''Ldapsearch wrapper'''
        rc, report = testlib.cmd(['ldapsearch', '-Y', 'DIGEST-MD5', '-H', 'ldap://127.0.0.1:389/', '-U', user, '-w', password, '-b', 'dc=example,dc=com', search])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        if exp_entries >= 1:
            self._word_find(report, "# numEntries: " + str(exp_entries))
        elif exp_entries == 0:
            name = "# numEntries: "
            warning = 'Found entries, but should not have: %s\n' % name
            self.assertFalse(name in report, warning + report)

    def test_daemons(self):
        '''(ServerSASL) Daemons running'''
        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

    def test_digest_md5(self):
        '''(ServerSASL) DIGEST-MD5 auth'''

        ServerCommon._initdb(self)

        # just store in openldap so we don't have to deal with sasldb2
        password = "{CLEARTEXT}foo"
        if self.release == "edgy" or self.release == "dapper":
            # workaround for http://www.openldap.org/its/index.cgi/Software%20Bugs?selectid=4021;statetype=-1
            password = "foo"

        ldif = '''#
dn: uid=sasluser,ou=Users,dc=example,dc=com
uid: sasluser
cn: Sasl User
objectClass: account
objectClass: posixAccount
objectClass: top
loginShell: /bin/bash
userPassword: ''' + password + '''
uidNumber: 5001
gidNumber: 5001
homeDirectory: /home/sasluser
gecos: Sasl User
description: Sasl User's account
'''
        self._modify(ldif, True)

        # can we bind via sasl?
        self._searchSASL('cn=*user', 'sasluser', 'foo', 2)

        # change password via sasl
        newpass = "bar"
        rc, report = testlib.cmd(['ldappasswd', '-Y', 'DIGEST-MD5', '-H', 'ldap://127.0.0.1:389/', '-U', 'sasluser', '-w', 'foo', '-s', newpass ])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

        # test new password via sasl
        self._searchSASL('cn=*user', 'sasluser', newpass, 2)

class ServerTLS(ServerCommon):
    '''Test OpenLDAP SSL and TLS.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)

        snake_pem = "/etc/ssl/certs/ssl-cert-snakeoil.pem"
        snake_key = "/etc/ssl/private/ssl-cert-snakeoil.key"
        
        subprocess.call(['cp', snake_pem, self.datadir])
        subprocess.call(['cp', snake_key, self.datadir])
        snake_pem = self.datadir + "/ssl-cert-snakeoil.pem"
        snake_key = self.datadir + "/ssl-cert-snakeoil.key"

        subprocess.call(['chown', self.uid + ':' + self.gid, snake_pem])
        subprocess.call(['chgrp', self.gid, snake_pem])
        subprocess.call(['chmod', '0440', snake_pem])
        subprocess.call(['chown', self.uid + ':' + self.gid, snake_key])
        subprocess.call(['chgrp', self.gid, snake_key])
        subprocess.call(['chmod', '0440', snake_key])

        subprocess.call(['sed', '-i', 's#^loglevel\(.*\)#loglevel 496\\1\\nTLSCipherSuite         HIGH\\nTLSCertificateFile     ' + snake_pem + '\\nTLSCertificateKeyFile  ' + snake_key + '#', self.config])

        subprocess.call(['sed', '-i', 's,^# SLAPD_SERVICES\(.*\),# SLAPD_SERVICES\\1\\nSLAPD_SERVICES="ldap:// ldaps://",', self.defaults])

        ServerCommon._verifyConfig(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def test_daemons(self):
        '''(ServerTLS) Daemons running'''
        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

    def test_tls(self):
        '''(ServerTLS) ldap:// (TLS)'''
        ServerCommon._initdb(self)

        try:
            os.chdir(self.datadir)
            fh = open(os.path.join(self.datadir, "ldaprc"), 'w')
            fh.write("TLS_REQCERT allow\n")
            fh.close()
        except:
            raise

        rc, report = testlib.cmd(['ldapsearch', '-ZZ', '-H', 'ldap://127.0.0.1:389/', '-D', "uid=testuser,ou=Users,dc=example,dc=com", '-w', 'pass', '-x', '-b', 'dc=example,dc=com', '(objectClass=posixAccount)' ])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        self._word_find(report, "# numEntries: 1")

        os.chdir("/tmp")        

    def test_ssl(self):
        '''(ServerTLS) ldaps:// (SSL)'''
        ServerCommon._initdb(self)

        try:
            os.chdir(self.datadir)        
            fh = open(os.path.join(self.datadir, "ldaprc"), 'w')
            fh.write("TLS_REQCERT allow\n")
            fh.close()
        except:
            raise

        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldaps://127.0.0.1:636/', '-D', "uid=testuser,ou=Users,dc=example,dc=com", '-w', 'pass', '-x', '-b', 'dc=example,dc=com', '(objectClass=posixAccount)' ])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        self._word_find(report, "# numEntries: 1")

        os.chdir("/tmp")        

class ServerIPC(ServerCommon):
    '''Test OpenLDAP IPC.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)

        subprocess.call(['sed', '-i', 's,^# SLAPD_SERVICES\(.*\),# SLAPD_SERVICES\\1\\nSLAPD_SERVICES="ldap:// ldapi://",', self.defaults])

        ServerCommon._verifyConfig(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def test_daemons(self):
        '''(ServerIPC) Daemons running'''
        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

    def test_ipc(self):
        '''(ServerIPC) ldapi:///'''
        ServerCommon._initdb(self)

        rc, report = testlib.cmd(['ldapsearch', '-H', 'ldapi:///', '-D', 'uid=testuser,ou=Users,dc=example,dc=com', '-w', 'pass', '-x', '-b', 'dc=example,dc=com', '(objectClass=posixAccount)'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        self._word_find(report, "# numEntries: 1")

class ServerCVEs(ServerCommon):
    '''Test OpenLDAP against CVEs.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)
        ServerCommon._restart(self)

        # populate with some entries
        ServerCommon._initdb(self)

    def tearDown(self):
        '''Shutdown methods'''
        ServerCommon._tearDown(self)

    def test_CVE_2007_5707(self):
        '''(ServerCVEs) CVE-2007-5707'''

        ldif = '''#
dn: uid=cve20075707,ou=Users,dc=example,dc=com
objectClasses: top
'''
        self._modify(ldif, True, True)

        # wait for it to actually die
        time.sleep(5)

        self.daemons = [ "slapd" ]
        ServerCommon._testDaemons(self, self.daemons)

class ServerOverlays(ServerCommon):
    '''Test OpenLDAP Overlay functionality.'''
    def _stop(self):
        '''Shutdown server'''
        subprocess.call([self.initscript, 'stop'], stdout=subprocess.PIPE)
        time.sleep(3)
        subprocess.call(['killall', 'slapd'], stdout=subprocess.PIPE,stderr=subprocess.PIPE)
        time.sleep(1)

    def _start(self):
        '''Start server'''
        rc, report = testlib.cmd([self.initscript, 'start'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

        if os.path.exists("/etc/ldap/slapd-pcache.conf"):
            rc, report = testlib.cmd(['slapd', '-g', self.gid, '-u', self.uid, '-f', "/etc/ldap/slapd-pcache.conf", '-4', '-l', 'USER', '-h', 'ldap://127.0.0.1:390/'])
            expected = 0
            result = 'Got exit code %d, expected %d\n' % (rc, expected)
            self.assertEquals(expected, rc, result + report)

        time.sleep(2)

    def _restart(self):
        self._stop()
        self._start()

    def _pcache_dbstat(self):
        '''Returns cache hits'''
        handle, name = tempfile.mkstemp(prefix='dbstatout',dir='/tmp')
        handle = file(name,'w')
        # if gutsy+n uses a different db version, need to change this
        subprocess.call(['db4.2_stat', '-m', '-h', self.pcachedbdir], stdout=handle)
        handle.close()

        fh = open(name, 'r')
        lines = fh.readlines()
        fh.close()
        os.unlink(name)

        line = ""
        found_id = False
        pat_id = re.compile(r'id2entry.bdb')
        for line in lines:
            if not found_id:
                if pat_id.search(line):
                    found_id = True
            else:
                if re.search(r'Requested pages found in the cache', line):
                    return int(line.split()[0])
        return -1


    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        self._stop()
        testlib.recursive_rm(self.pcachedbdir)
        os.unlink(self.pcacheconf)
        ServerCommon._tearDown(self)

    def test_pcache(self):
        '''(ServerOverlays) Proxy and ProxyCache'''

        self.pcachedbdir = "/var/lib/ldap-pcache"
        self.pcacheconf = "/etc/ldap/slapd-pcache.conf"

        # release specific config
        pcache_name = "pcache"
	added_modules = '''#
moduleload      back_ldap
'''
        if self.release == "edgy" or self.release == "dapper":
            pcache_name = "proxycache"
            added_modules = ""

        # pcache config
        pconf = '''#
allow bind_v2
include         /etc/ldap/schema/core.schema
include         /etc/ldap/schema/cosine.schema
include         /etc/ldap/schema/nis.schema
include         /etc/ldap/schema/inetorgperson.schema
pidfile         ''' + self.rundir + '''/slapd-pcache.pid
argsfile        ''' + self.rundir + '''/slapd-pcache.args
loglevel        200
modulepath      /usr/lib/ldap
moduleload      back_bdb
moduleload      pcache
''' + added_modules + '''
sizelimit       500
tool-threads    1
database        ldap
suffix          "dc=example,dc=com"
rootdn          "dc=example,dc=com"
uri             "ldap://127.0.0.1:389/dc=example%2cdc=com"
overlay         ''' + pcache_name + '''
proxycache      bdb 100000 1 1000 100
# for some reason openldap 2.2 likes only one attr in proxyAttrset
proxyAttrset    0 uidNumber
proxyTemplate   (cn=) 0 3600
proxyTemplate   (uid=) 0 3600
proxyTemplate   (&(objectClass=)(cn=)) 0 3600
proxyTemplate   (&(objectClass=)(uid=)) 0 3600
cachesize       20
directory       ''' + self.pcachedbdir + '''
dbconfig set_cachesize 0 2097152 0
index       objectClass eq
index       cn,uid  pres,eq,sub
'''
        try:
            fh = open(self.pcacheconf, 'w')
            fh.write(pconf)
            fh.close()
        except:
            raise
       
        os.makedirs(self.pcachedbdir)
        subprocess.call(['chown', self.uid + ':' + self.gid, self.pcachedbdir])

        self._restart()

        # check both are running
        for d in ['slapd', 'slapd-pcache']:
            pidfile = os.path.join(self.rundir, d + ".pid")
            warning = "Could not find pidfile '" + pidfile + "'"
            self.assertTrue(os.path.exists(pidfile), warning)
            self.assertTrue(testlib.check_pid("slapd", pidfile))

        # populate initially
        ServerCommon._initdb(self)

        hits = self._pcache_dbstat()
        assert hits >= 0, "Bad db stats"

        # test the proxy
        rc, report = testlib.cmd(['ldapsearch', '-x', '-H', 'ldap://127.0.0.1:390/', '-b', 'dc=example,dc=com', '(&(objectClass=posixAccount)(uid=testuser))', 'uidNumber'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        self._word_find(report, "# numEntries: 1")

        newhits = self._pcache_dbstat()
        assert newhits > hits, "Cache miss '" + str(newhits) + " < " + str(hits) + "'"


class ServerStub(ServerCommon):
    '''Debugging stub.'''
    def setUp(self):
        '''Setup mechanisms'''
        ServerCommon._setUp(self)
        ServerCommon._restart(self)

    def tearDown(self):
        '''Shutdown methods'''
        #ServerCommon._tearDown(self)
        pass

    def test_stub(self):
        '''(ServerStub) test_stub'''
        ServerCommon._initdb(self)

#
# MAIN
#
if __name__ == '__main__':
    assert not os.path.isdir("/var/lib/ldap.autotest")
    subprocess.call(['mv', "/var/lib/ldap", "/var/lib/ldap.autotest"])

    suite = unittest.TestSuite()

    # add tests here
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerGeneric))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerSimple))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerSimpleHDB))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerIPC))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerSASL))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerTLS))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerOverlays))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerCVEs))

    # only use for debugging-- it doesn't cleanup
    #suite.addTest(unittest.TestLoader().loadTestsFromTestCase(ServerStub))

    # run tests
    unittest.TextTestRunner(verbosity=2).run(suite)

    # make sure slapd isn't running
    print "Killing stray slapd processes"
    subprocess.call(['killall', 'slapd'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    subprocess.call(['killall', 'slurd'], stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    testlib.recursive_rm("/var/lib/ldap")
    subprocess.call(['mv', "/var/lib/ldap.autotest", "/var/lib/ldap"])
