#!/usr/bin/python
'''
    How to run against a clean schroot named 'feisty':
        schroot -c feisty -u root -- sh -c 'apt-get -y install python-unit php5-cli php5-sqlite php5-gd && ./test-php5.py -v'

'''

import unittest, subprocess, shutil, grp, os, os.path, sys, time
import testlib

class PHP5Test(testlib.TestlibCase):
    '''Test php5 functionality.'''

    def _run_script(self, contents, expected=0, args=[]):
        '''Run a php script, expecting exit code 0'''
        handle, name = testlib.mkstemp_fill('<?php\n'+contents+'\n?>\n')
        self.assertShellExitEquals(expected, ['/usr/bin/php5'] + args, stdin = handle)
        os.unlink(name)

    def test_a_simple_application(self):
        '''Simple "Hello World" application'''

        self._run_script('echo "Hello World";')

    def test_mopb_10(self):
        '''Protected against MOPB 10'''

        self._run_script('''
  session_start();
  $x = chr(36).str_repeat("A", 36)."N;".chr(127);
  $data = $x;
  
  session_decode($data);
  $keys = array_keys($_SESSION);
  $heapdump = $keys[1];

  exit(strlen($heapdump));
''', args=['-d','session.serialize_handler=php_binary'])

    def test_mopb_14(self):
        '''Protected against MOPB 14'''

        self._run_script('''
  $sizeofHashtable = 39;
  $maxlong = 0x7fffffff;
  if (is_int($maxlong+1)) {
    $sizeofHashtable = 67;
    $maxlong = 0x7fffffffffffffff;
  }

  $memdump = str_repeat("A", 4096);
  for ($i=0; $i<40; $i++) $d[] = array();
  unset($d[20]);
  $x = str_repeat("A", $sizeofHashtable);
  
  // If the libc memcmp leaks the information use it
  // otherwise we only get a case insensitive memdump
  $b = substr_compare(chr(65),chr(0),0,1,false) != 65;

  for ($i=0; $i<4096; $i++) {
    $y = substr_compare($x, chr(0), $i+1, $maxlong, $b);
    $Y = substr_compare($x, chr(1), $i+1, $maxlong, $b);
    if ($y-$Y == 1 || $Y-$y==1){
      $y = chr($y);
      if ($b && strtoupper($y)!=$y) {
        if (substr_compare($x, $y, $i+1, $maxlong, false)==-1) {
          $y = strtoupper($y);
        }
      }
      $memdump[$i] = $y;
    } else {
      $memdump[$i] = chr(0);
    }
  }

  $rc = 0;
  for ($b=0; $b<strlen($memdump); $b+=1) {
    if ($memdump[$b] != chr(0)) {
        $rc = 1;
    }
  }
  if ($rc) {
    echo "Saw leaked memory in memdump\n";
  }

  exit($rc);
  
''')

    def test_mopb_15(self):
        '''Protected against MOPB 15'''

        self._run_script('''
  function init()
  {
    global $rid;
    
    $rid = imagecreate(10,10);
    imagecolorallocate($rid, 0, 0, 0);
    imagecolorallocate($rid, 0, 0, 0);
  }
  
  function peek($addr, $size)
  {
    global $rid;
    imagecolordeallocate($rid, 0);
    imagecolordeallocate($rid, 1);
    imagecolorallocate($rid, $addr, 0, 0);
    imagecolorallocate($rid, $size, 0, 0);
    return shmop_read((int)$rid, 0, $size);
  }

  init();

    $maps = fopen("/proc/self/maps","r");
    while (!feof($maps)) {
        $line = fgets($maps, 4096);
        echo $line;
        $sections = explode(' ',$line);
        $addrs = explode('-',$sections[0]);
        $perms = $sections[1];

        $offset = intval($addrs[0],16);
        $size = intval($addrs[1],16) - $offset;

        if ($size && strpos($perms,"r")!==false && strpos($perms,"w")!==false) {
            $span = $size;
            if ($span > 1024) {
                $span = 1024;
            }

            echo "Attempting to read $span from $offset\n";

            $data = peek($offset,$span);
            if (strlen($data) == $span) {
                echo "Oops, able to read memory!\n";
                exit(1);
            }
            echo "Unable to read memory\n";
            exit(0);
        }
    }
''')

    def test_mopb_22(self):
        '''Protected against MOPB 22'''

        self._run_script('''
  $offset_1 = 0x55555555;
  $offset_2 = 0x08048040;
  // Convert offsets into strings
  $addr1 = pack("L", $offset_1);
  $addr2 = pack("L", $offset_2);

  define("C0", $addr1[0]);
  define("C1", $addr1[1]);
  define("C2", $addr1[2]);
  define("C3", $addr1[3]);
  
  define("M0", $addr2[0]);
  define("M1", $addr2[1]);
  define("M2", $addr2[2]);
  define("M3", $addr2[3]);
  
  function myErrorHandler()
  {
    session_id(str_repeat("A", 100));
    
    $GLOBALS['str'] = str_repeat("A", 39);
    
    for ($i=0; $i<7; $i++) {
      $GLOBALS['str'][$i*4+0] = M0;
      $GLOBALS['str'][$i*4+1] = M1;
      $GLOBALS['str'][$i*4+2] = M2;
      $GLOBALS['str'][$i*4+3] = M3;
    }
    $GLOBALS['str'][8*4+0] = C0;
    $GLOBALS['str'][8*4+1] = C1;
    $GLOBALS['str'][8*4+2] = C2;
    $GLOBALS['str'][8*4+3] = C3;

    return true;
  }
  
  function doit()
  {
    error_reporting(E_ALL);
    set_error_handler("myErrorHandler");
    session_id(str_repeat("A", 39));
    session_start();
    session_regenerate_id();
  }
  
  doit();
''',args=['-d','session.hash_bits_per_character=666'])

    def test_mopb_24(self):
        '''Protected against MOPB 24'''

        self._run_script('''
  $shellcode = str_repeat(chr(0x90), 1000);

  $arr = array(str_repeat("A", 39) => 1, "B" => 1);

  function array_compare(&$key1, &$key2)
  {
    $GLOBALS['a'] = &$key2;
    unset($key2);
    return 1;
  }

  uksort($arr, "array_compare");
  $x=array($shellcode => 1);

  $a[8*4+0] = $a[6*4+0];
  $a[8*4+1] = chr(ord($a[6*4+1])+2); // <--- This only works for Little Endian
  $a[8*4+2] = $a[6*4+2];
  $a[8*4+3] = $a[6*4+3];

  unset($x);
''')

    # This needs to use mod_php to test for long-running register_globals
    # behavior.
    def _test_mopb_26(self):
        '''Protected against MOPB 26'''

        self._run_script('''
  $str = "a=".str_repeat("A", 164000);

  // This code just fills the memory up to the limit...
  $limit = ini_get("memory_limit");
  if (strpos($limit, "M")) {
    $limit *= 1024 * 1024;
  } else if (strpos($limit, "K")) {
    $limit *= 1024;
  } else $limit *=1;
  while ($limit - memory_get_usage(true) > 2048) $x[] = str_repeat("A", 1024);

  // Will activate register_globals and trigger the memory_limit
  mb_parse_str($str);
''')

    def test_mopb_29(self):
        '''Protected against MOPB 29'''

        self._run_script('''
    function checkit($length) {
        if ($length == 0 || $length == 100) {
            echo "Okay ($length)\\n";
            exit(0);
        }
        echo "Bad ($length)\\n";
        exit(1);
    }

  $str = 'S:'.(100*3).':"'.str_repeat('\\61', 100).'"';
  $arr = array(str_repeat('"', 200)."1"=>1,str_repeat('"', 200)."2"=>1);

  $heapdump = unserialize($str);
  checkit(strlen($heapdump));

''')

    def test_mopb_30(self):
        '''Protected against MOPB 30'''

        self._run_script('''
  $shellcode = str_repeat("\\x90", 400);
  $zend_execute_internal = 0x90909090;

  $Hashtable = pack("LLLLLLLLLCCC", 2, 1, 0, 0, 0, $zend_execute_internal, 0, $zend_execute_internal, 0x66666666, 0, 0, 0);

  eval('
  session_start();
  unset($HTTP_SESSION_VARS);
  unset($_SESSION);
  $x = "'.$Hashtable.'";
  session_register($shellcode);');
''')

    def test_mopb_34(self):
        '''Protected against MOPB 34'''

        self._run_script('''
if (mail("test@example.com", "Test\r\n \nAnother-Header: Blub", "Message")===FALSE) {
    echo "Email had injected headers visible\n";
    exit(1);
}
echo "Email was correctly missing injected headers\n";
exit(0);
#''', args=['-d','sendmail_path=bash -c "grep -q ^Another-Header && exit 1 || exit 0"'])

    def test_mopb_41(self):
        '''Protected against MOPB 41'''

        self._run_script('''
    $z = "UUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUUU";
    $y = "DDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD";
    $x = "AQ                                                                        ";
    unset($z);
    unset($y);
    $x = base64_decode($x);

    $y = sqlite_udf_decode_binary($x);

    unset($x);
''')

    def test_mopb_45(self):
        '''Protected against MOPB 45'''

        self._run_script('''
  if (version_compare(phpversion(), "5.2.0")==-1) {
    echo "filter_var does not exist\n";
    exit(0);
  }

  $orig = "test@example.com\n";
  $final = filter_var($orig, FILTER_VALIDATE_EMAIL);

  if ($final === FALSE) {
    echo "Filter detected EOL newline\n";
    exit(0);
  }
  if ($orig == $final) {
    echo "Filter failed to see EOL newline\n";
    exit(1);
  }
  echo "Something odd happened to the filter: '$final' != 'test@example.com'\n";
  exit(1);
''')


if __name__ == '__main__':
    unittest.main()
