#!/usr/bin/python
'''
    How to run against a clean schroot named 'feisty':
        schroot -c feisty -u root -- sh -c 'apt-get -y install python-unit samba smbclient && ./test-samba.py -v'
'''

import unittest, subprocess, tempfile, os, os.path, grp, time
import testlib

class SambaCommon(testlib.TestlibCase):
    '''Common routines for testing samba shares.'''
    def _restart(self):
        '''Shutdown and startup with each test'''
        subprocess.call(['/etc/init.d/samba', 'stop'], stdout=subprocess.PIPE)
        self.assertTrue(subprocess.call(['/etc/init.d/samba', 'start'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT) == 0)
	time.sleep(3)

    def _browse(self):
        '''Issue a browse request and check for a specific string'''
        rc, report = testlib.cmd(['smbclient','-N','-L','localhost'])
        expected = 0
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)
        return report

    def _smbcmd(self,mount,cmd,exitcode=0,user=None,password=None):
        smbclient = ['smbclient']
        if password==None:
            smbclient += ['-N']
        if user!=None:
            smbclient += ['-U',user]
        smbclient += ['-c',cmd,'//localhost'+mount]
        if password!=None:
            smbclient += [password]

        rc, report = testlib.cmd(smbclient)
        result = 'Got exit code %d, expected %d\nCommand: %s\n' % (rc, exitcode, ' '.join(smbclient))
        self.assertEquals(exitcode, rc, result + report)

        return report

    def _smbget(self,mount,path,exitcode=0,user=None,password=None):
        '''Get contents of a remote file'''
        dir = os.path.dirname(path) 
        base = os.path.basename(path)

        tmpdir = tempfile.mkdtemp(dir='/tmp')
        cmd = 'lcd '+tmpdir+';get '+base

        report = self._smbcmd(mount,cmd,exitcode,user,password)

        contents = ''
        if exitcode == 0:
            localpath = os.path.join(tmpdir,base)
            contents = file(localpath).read()
            os.unlink(localpath)

        os.rmdir(tmpdir)
        return contents, report

    def _smbput(self,local_path,mount,target_dir,exitcode=0,user=None,password=None):
        '''Put local file to remote path'''
        local_dir = os.path.dirname(local_path) 
        local_base = os.path.basename(local_path)

        cmd = 'lcd '+local_dir+';cd '+target_dir+';put '+local_base
        return self._smbcmd(mount,cmd,exitcode,user,password)

    def _smbdir(self,mount,target_dir,exitcode=0,user=None,password=None):
        '''Get directory of remote path'''

        cmd = 'cd '+target_dir+';dir'
        return self._smbcmd(mount,cmd,exitcode,user,password)

    def _word_find(self,report,name):
        '''Check for a specific string'''
        warning = 'Could not find "%s"\n' % name
        self.assertTrue(name in report, warning + report)

class SambaGenericTest(SambaCommon):
    '''Test for generic anonymous functions'''
    def setUp(self):
        '''Generic test setup'''
        SambaCommon._restart(self)

    def test_browse_ipc(self):
        '''Can browse IPC share and Workgroup'''
        report = self._browse()
        self._word_find(report,'IPC$')
        self._word_find(report,'MSHOME')
        self._word_find(report,'Samba, Ubuntu')
        self._word_find(report,'Anonymous login successful')

    def test_browse_only_printdriver_disk(self):
        '''Initially only a print driver share'''
        disks = [f for f in self._browse().splitlines() if ' Disk ' in f]
        for f in disks:
            self.assertTrue('print$' in f,'Unexpected "Disk" share:\n' + f)

class SambaTmpTest(SambaCommon):
    '''Test a world-writable share'''

    def setUp(self):
        '''Tmp test setup'''
        SambaCommon._restart(self)

    def test_tmp_00_initialize(self):
        '''Tmp share: set up configuration'''
        testlib.config_replace('/etc/samba/smb.conf','''#
[tmp]
   comment = Temp Directory
   path = /tmp
   browseable = yes
   writable = yes
   public = yes
   guest ok = yes
''',append=True)

        SambaCommon._restart(self)

    def test_tmp_browse(self):
        '''Tmp share: browseable'''
        report = self._browse()
        disks = [f for f in report.splitlines() if ' Disk ' in f]
        found = False
        for f in disks:
            if 'Temp Directory' in f:
                found = True
        self.assertTrue(found, 'Did not find tmp share:\n' + report)

    def test_tmp_file_read(self):
        '''Tmp share: read remote file contents'''
        string = 'This is a test for reading'
        handle, name = tempfile.mkstemp(prefix='test-read',dir='/tmp')
        handle = file(name,'w')
        handle.write(string)
        handle.close()

        # verify that we cannot read a private file on an anon share
        contents, report = self._smbget('/tmp',name, exitcode=1)
        self.assertTrue('NT_STATUS_ACCESS_DENIED opening remote file' in report, 'Should not be able to read file:\n' + report)

        os.chmod(name,0744)
        contents, report = self._smbget('/tmp',name)
        self.assertTrue('getting file ' in report, 'Should be able to read file:\n' + report)

        self.assertTrue(string == contents, 'Contents miss-match "%s" != "%s"' % (string,contents))

        os.unlink(name)

    def test_tmp_file_write(self):
        '''Tmp share: write remote file contents'''
        string = 'This is a test for writing'
        handle, name = tempfile.mkstemp(prefix='test-write',dir='/tmp')
        handle = file(name,'w')
        handle.write(string)
        handle.close()

        tmpdir = tempfile.mkdtemp(prefix='target-dir',dir='/tmp')
        # Make sure we can't write to private directories
        report = self._smbput(name,'/tmp',os.path.basename(tmpdir),exitcode=1)
        self.assertTrue('Failed to open ' in report or 'NT_STATUS_ACCESS_DENIED opening remote file ' in report, 'Should not be able to write file:\n' + report)

        os.chmod(tmpdir,0777)
        report = self._smbput(name,'/tmp',os.path.basename(tmpdir))
        self.assertTrue('putting file ' in report, 'Should be able to write file:\n' + report)

        newpath = os.path.join(tmpdir,os.path.basename(name))
        contents = file(newpath).read()
        self.assertTrue(string == contents, 'Contents miss-match "%s" != "%s"' % (string,contents))

        os.unlink(name);
        os.unlink(newpath);
        os.rmdir(tmpdir);

    def test_tmp_zz_cleanup(self):
        '''Tmp share: clean up configuration'''
        testlib.config_restore('/etc/samba/smb.conf')

        SambaCommon._restart(self)


class SambaUserTest(SambaCommon):
    '''Test access to a user-writable share'''

    def setUp(self):
        '''User test setup'''
        SambaCommon._restart(self)

        self.user = testlib.TestUser()#group='users',uidmin=2000,lower=True)

        subprocess.call(['groupadd','evilsamba'])
        subprocess.call(['usermod','-G','users,evilsamba',self.user.login])

        if not os.path.exists('/var/tmp/samba-forcedgroup'):
            os.mkdir('/var/tmp/samba-forcedgroup', 0755)
        if not os.path.exists('/var/tmp/samba-forcedgroup/readable'):
            os.mkdir('/var/tmp/samba-forcedgroup/readable')
        if not os.path.exists('/var/tmp/samba-forcedgroup/writable'):
            os.mkdir('/var/tmp/samba-forcedgroup/writable')
        if not os.path.exists('/var/tmp/samba-forcedgroup/unreadable'):
            os.mkdir('/var/tmp/samba-forcedgroup/unreadable')
        os.chown('/var/tmp/samba-forcedgroup/readable', 0, grp.getgrnam('users')[2])
        os.chmod('/var/tmp/samba-forcedgroup/readable', 0770)
        os.chown('/var/tmp/samba-forcedgroup/writable', 0, grp.getgrnam('evilsamba')[2])
        os.chmod('/var/tmp/samba-forcedgroup/writable', 0770)
        os.chown('/var/tmp/samba-forcedgroup/unreadable', 0, 0)
        os.chmod('/var/tmp/samba-forcedgroup/unreadable', 0770)

        #subprocess.call(['ls','-la','/var/tmp/samba-forcedgroup'])
        #subprocess.call(['id',self.user.login])

        testlib.config_replace('/etc/samba/smb.conf','''#
[forcedgroup]
   comment = Test Directory
   path = /var/tmp/samba-forcedgroup
   create mode = 0664
   directory mode = 2775
   force group = evilsamba
   valid users = @users
   writable = yes
''',append=True)

        # smbpasswd seems to collide on its auth files randomly.
        for i in ['account_policy','group_mapping','passdb','secrets']:
            tdb = os.path.join('/var/lib/samba',i+'.tdb')
            if os.path.exists(tdb):
                os.unlink(tdb)

        SambaCommon._restart(self)

        # Set SMB password
        prompt, name = testlib.mkstemp_fill(self.user.password+'\n'+self.user.password+'\n')
        os.unlink(name)
        self.assertShellExitEquals(0,['smbpasswd','-L', '-s','-a',self.user.login], stdin = prompt)

        # Enable SMB user
        self.assertShellExitEquals(0,['smbpasswd','-L', '-e',self.user.login])

	# needed for Dapper and Edgy
        SambaCommon._restart(self)

    def tearDown(self):
        '''User test cleanup'''

        #print 'check it out.... %s %s' % (self.user.login,self.user.password)
        #subprocess.call(['bash'])

        os.rmdir('/var/tmp/samba-forcedgroup/writable')
        os.rmdir('/var/tmp/samba-forcedgroup/readable')
        os.rmdir('/var/tmp/samba-forcedgroup/unreadable')
        os.rmdir('/var/tmp/samba-forcedgroup')

        #if self.user != None:
        #    self.assertShellExitEquals(0,['smbpasswd','-L', '-x',self.user.login])
        testlib.config_restore('/etc/samba/smb.conf')
        subprocess.call(['groupdel','evilsamba'])
        self.user = None

    def test_user_dir_readable(self):
        '''User share: readable directory contents'''

        # This directory should be unreadable
        report = self._smbdir('/forcedgroup','unreadable',user=self.user.login, password=self.user.password)
        self.assertTrue('NT_STATUS_ACCESS_DENIED listing ' in report, 'Should not be able to read directory:\n' + report)

        # This directory should be readable
        report = self._smbdir('/forcedgroup','readable',user=self.user.login, password=self.user.password)
        self.assertTrue(' D        0 ' in report, 'Should be able to read directory:\n' + report)

    def test_user_file_write(self):
        '''User share: write remote file contents'''
        string = 'This is a test for writing with group perms'
        handle, name = tempfile.mkstemp(prefix='test-write',dir='/tmp')
        handle = file(name,'w')
        handle.write(string)
        handle.close()

        # This should not be writable
        report = self._smbput(name,'/forcedgroup','.',user=self.user.login, password=self.user.password,exitcode=1)
        self.assertTrue('Failed to open ' in report or 'NT_STATUS_ACCESS_DENIED opening remote file ' in report, 'Should not be able to write file:\n' + report)

        # This should be writable
        report = self._smbput(name,'/forcedgroup','writable',user=self.user.login, password=self.user.password)
        self.assertTrue('putting file ' in report, 'Should be able to write file:\n' + report)

        newpath = os.path.join('/var/tmp/samba-forcedgroup/writable',os.path.basename(name))
        contents = file(newpath).read()
        self.assertTrue(string == contents, 'Contents miss-match "%s" != "%s"' % (string,contents))

        os.unlink(name);
        os.unlink(newpath);


if __name__ == '__main__':
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(SambaGenericTest))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(SambaTmpTest))
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(SambaUserTest))
    unittest.TextTestRunner(verbosity=2).run(suite)

    #print "Shutting down samba"
    subprocess.call(['/etc/init.d/samba', 'stop'], stdout=subprocess.PIPE)

