'''Common classes and functions for package tests.'''

import string, random, crypt, subprocess, pwd, signal, time, unittest, tempfile, shutil, os, os.path, re
from stat import *

class TimedOutException(Exception):
    def __init__(self, value = "Timed Out"):
        self.value = value
    def __str__(self):
        return repr(self.value)

def config_replace(path,contents,append=False):
    '''Replace (or append) to a config file'''
    pathbackup = path+ '.autotest'
    if os.path.exists(pathbackup):
        config_restore(path)
    if os.path.exists(path):
        if not os.path.exists(pathbackup):
            shutil.copyfile(path, pathbackup)
        if append:
            contents = file(path).read() + contents
    open(path, 'w').write(contents)

def config_patch(path,patch):
    '''Patch a config file'''
    pathbackup = path+ '.autotest'
    if os.path.exists(pathbackup):
        config_restore(path)
    if os.path.exists(path) and not os.path.exists(pathbackup):
        shutil.copyfile(path, pathbackup)

    handle, name = mkstemp_fill(patch)
    rc = subprocess.call(['/usr/bin/patch', '-p1', path], stdin=handle, stdout=subprocess.PIPE)
    os.unlink(name)
    if rc != 0:
        raise Exception("Patch failed")
        

def config_restore(path):
    '''Rename a replaced config file back to its initial state'''
    pathbackup = path+ '.autotest'
    if os.path.exists(pathbackup):
        os.rename(pathbackup, path)

def timeout(secs, f, *args):
    def handler(signum, frame):
        raise TimedOutException()
    
    old = signal.signal(signal.SIGALRM, handler)
    result = None
    signal.alarm(secs)
    try:
        result = f(*args)
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, old)

    return result

def random_string(length,lower=False):
    '''Return a random string, consisting of ASCII letters, with given
    length.'''

    s = ''
    selection = string.letters
    if lower:
        selection = string.lowercase
    maxind = len(selection)-1
    for l in range(length):
        s += selection[random.randint(0, maxind)]
    return s

def mkstemp_fill(contents,suffix='',prefix='testlib-',dir=None):
    '''As tempfile.mkstemp does, return a (file, name) pair, but with
    prefilled contents.'''

    handle, name = tempfile.mkstemp(suffix=suffix,prefix=prefix,dir=dir)
    handle = file(name,"w+")
    handle.write(contents)
    handle.flush()
    handle.seek(0)
    
    return handle, name

def login_exists(login):
    '''Checks whether the given login exists on the system.'''

    try:
        pwd.getpwnam(login)
        return True
    except KeyError:
        return False

def recursive_rm(dirPath):
    '''recursively remove directory'''
    names = os.listdir(dirPath)
    for name in names:
        path = os.path.join(dirPath, name)
        if not os.path.isdir(path):
            os.unlink(path)
        else:
            recursive_rm(path)
    os.rmdir(dirPath)

def check_pid(exe, pidfile):
    '''Checks if pid in pidfile is running'''
    if not os.path.exists(pidfile):
        return False

    # get the pid
    try:
        fd = open(pidfile, 'r')
        pid = fd.readline().rstrip('\n')
        fd.close()
    except:
        return False

    cmdline = "/proc/" + pid + "/cmdline"
    if not os.path.exists(cmdline):
        return False

    # get the command line
    try:
        fd = open(cmdline, 'r')
        tmp = fd.readline().split('\0')
        fd.close()
    except:
        return False

    # this allows us to match absolute paths or just the executable name
    if re.match('^' + exe + '$', tmp[0]) or re.match('.*/' + exe + '$', tmp[0]):
        return True

    return False

def ubuntu_release():
    f = "/etc/lsb-release"
    try:
        size = os.stat(f)[ST_SIZE]
    except:
        raise

    if size > 1024*1024:
        raise IOError, 'Could not open "%s" (too big)' % f

    try:
        fh = open("/etc/lsb-release", 'r')
    except:
        raise

    lines = fh.readlines()
    fh.close()

    pat = re.compile(r'DISTRIB_CODENAME')
    for line in lines:
        if pat.search(line):
            return line.split('=')[1].rstrip('\n').rstrip('\r')

    return "UNKNOWN"

def cmd(command, input = None, stderr = subprocess.STDOUT, stdout = subprocess.PIPE, stdin = None):
    '''Try to execute given command (array) and return its stdout, or return
    a textual error if it failed.'''

    try:
       sp = subprocess.Popen(command, stdin=stdin, stdout=stdout, stderr=stderr, close_fds=True)
    except OSError, e:
        return [127, str(e)]

    out = sp.communicate(input)[0]
    return [sp.returncode,out]

class TestlibCase(unittest.TestCase):

    def _testlib_shell_cmd(self,args,stdin=None):
        rc, out = cmd(args,stdin=stdin)
        argstr = "'" + "', '".join(args).strip() + "'"
        report = 'Command: ' + argstr + '\nOutput:\n' + out
        return rc, report

    def assertShellExitEquals(self, expected, args, stdin=None):
        '''Test a shell command matches a specific exit code'''
        rc, report = self._testlib_shell_cmd(args, stdin=stdin)
        result = 'Got exit code %d, expected %d\n' % (rc, expected)
        self.assertEquals(expected, rc, result + report)

    def assertShellExitNotEquals(self, unwanted, args, stdin=None):
        '''Test a shell command doesn't match a specific exit code'''
        rc, report = self._testlib_shell_cmd(args, stdin=stdin)
        result = 'Got (unwanted) exit code %d\n' % rc
        self.assertNotEquals(unwanted, rc, result + report)

class TestUser:
    '''Create a temporary test user and remove it again in the dtor.'''

    def __init__(self, login=None, home=True, group=None, uidmin=None, lower=False):
        '''Create a new user account with a random password. 
        
        By default, the login name is random, too, but can be explicitly
        specified with 'login'. By default, a home directory is created, this
        can be suppressed with 'home=False'.'''

        self.login = None

        if login:
            if login_exists(login):
                raise ValueError, 'login name already exists'
        else:
            while(True):
                login = random_string(8,lower=lower)
                if not login_exists(login):
                    break

        self.salt = random_string(2)
        self.password = random_string(8,lower=lower)
        self.crypted = crypt.crypt(self.password, self.salt)

        creation = ['useradd', '-p', self.crypted]
        if home:
            creation += ['-m']
        if group:
            creation += ['-G',group]
        if uidmin:
            creation += ['-K','UID_MIN=%d'%uidmin]
        creation += [login]
        assert subprocess.call(creation) == 0

        self.login = login
        p = pwd.getpwnam(self.login)
        self.uid = p[2]
        self.gid = p[3]

    def __del__(self):
        '''Remove the created user account.'''

        if self.login:
            # seems to already have gone here
            try:
                import subprocess
            except:
                pass
            assert subprocess.call(['userdel', '-r', self.login]) == 0
